import 'dart:math';

import 'package:fl_chart/fl_chart.dart';
import 'package:flutter/material.dart';
import 'package:intl/intl.dart';
import 'package:og1_chart/src/enums/_enums.dart';
import 'package:og1_chart/src/models/_models.dart';

class Og1LineChartService {
  FlTitlesData createTitlesData(
      String? yAxisMetric,
      Color? titleColor,
      double? xTitlesFontSize,
      double? yTitlesFontSize,
      double? xInterval,
      double? yInterval) {
    return FlTitlesData(
        show: true,
        bottomTitles: AxisTitles(
            sideTitles:
                createBottomTitles(titleColor, xTitlesFontSize, xInterval)),
        leftTitles: AxisTitles(
            sideTitles: createLeftTitles(
                yAxisMetric, titleColor, yTitlesFontSize, yInterval)),
        rightTitles: AxisTitles(sideTitles: SideTitles(showTitles: false)),
        topTitles: AxisTitles(sideTitles: SideTitles(showTitles: false)));
  }

  SideTitles createBottomTitles(
      Color? titleColor, double? xTitlesFontSize, double? xInterval) {
    return SideTitles(
      showTitles: true,
      reservedSize: 35,
      interval: xInterval,
      getTitlesWidget: (value, meta) => Container(
        margin: const EdgeInsets.all(8),
        child: Text(xAxisTitles(value, null),
            style: TextStyle(
              color: titleColor ?? const Color(0xff68737d),
              fontWeight: FontWeight.bold,
              fontSize: xTitlesFontSize ?? 10,
            )),
      ),
    );
  }

  String xAxisTitles(double xAxisValue, XAxisEnum? xAxisEnum) {
    xAxisEnum ??= XAxisEnum.month;
    List<String> xAxis = xAxisString(xAxisEnum);
    return xAxis.asMap().containsKey(_roundDouble(xAxisValue, 5).toInt())
        ? xAxis[_roundDouble(xAxisValue, 5).toInt()]
        : '';
  }

  double _roundDouble(double value, int places) {
    num mod = pow(10.0, places);
    return ((value * mod).roundToDouble() / mod);
  }

  SideTitles createLeftTitles(String? yAxisMetric, Color? titleColor,
      double? yTitlesFontSize, double? yInterval) {
    return SideTitles(
      showTitles: true,
      reservedSize: 60,
      interval: yInterval,
      getTitlesWidget: (value, meta) => Container(
        margin: const EdgeInsets.all(12),
        child: Text(yAxisTitles(value, yAxisMetric),
            style: TextStyle(
              color: titleColor ?? const Color(0xff67727d),
              fontWeight: FontWeight.bold,
              fontSize: yTitlesFontSize ?? 10,
            )),
      ),
    );
  }

  String yAxisTitles(double yAxisValue, String? yAxisMetric) {
    yAxisValue = yAxisValue.roundToDouble();
    yAxisMetric ??= '';
    return "$yAxisValue $yAxisMetric".trim();
  }

  LineChartBarData createLineChartBarData(
      ChartLine chartLine, XAxisEnum xAxisEnum) {
    List<ChartLineSpot> avgChartLineSpots =
        avg(chartLine.chartLineSpots, xAxisEnum);
    List<ChartLineSpot> meanChartLineSpots = mean(avgChartLineSpots);
    List<FlSpot> spots = flSpots(meanChartLineSpots, xAxisEnum);

    return LineChartBarData(
        isCurved: true,
        color: chartLine.lineColors.elementAt(0),
        barWidth: chartLine.barWidth ?? 5,
        isStrokeCapRound: true,
        dotData: FlDotData(show: false),
        belowBarData: BarAreaData(show: false),
        spots: spots);
  }

  List<DateTime> xAxisDateTime(XAxisEnum xAxisEnum) {
    DateFormat dateFormat =
        xAxisEnum == XAxisEnum.month ? DateFormat('MMM') : DateFormat('YYYY');
    DateTime now = DateTime.now();
    dateFormat.format(now);
    return xAxisEnum == XAxisEnum.month
        ? [
            DateTime(now.year, now.month - 3),
            DateTime(now.year, now.month - 2),
            DateTime(now.year, now.month - 1),
            DateTime(now.year, now.month)
          ]
        : [
            DateTime(now.year - 3),
            DateTime(now.year - 2),
            DateTime(now.year - 1),
            DateTime(now.year)
          ];
  }

  List<String> xAxisString(XAxisEnum xAxisEnum) {
    DateFormat dateFormat =
        xAxisEnum == XAxisEnum.month ? DateFormat('MMM') : DateFormat('YYYY');
    return xAxisDateTime(xAxisEnum)
        .map((date) => dateFormat.format(date))
        .toList();
  }

  List<ChartLineSpot> avg(
      List<ChartLineSpot> chartLineSpots, XAxisEnum xAxisEnum) {
    DateFormat dateFormat =
        xAxisEnum == XAxisEnum.month ? DateFormat('MMM') : DateFormat('YYYY');
    List<DateTime> lastFour = xAxisDateTime(xAxisEnum);
    List<String> lastFourByX = xAxisString(xAxisEnum);

    chartLineSpots
        .where((chartLineSpot) =>
            lastFourByX.contains(dateFormat.format(chartLineSpot.date)))
        .toList()
        .sort((a, b) => a.date.compareTo(b.date));

    List<ChartLineSpot> avgChartLineSpots = lastFour
        .map((xAxis) => avgSpot(chartLineSpots, xAxis, dateFormat))
        .toList();

    return avgChartLineSpots;
  }

  ChartLineSpot avgSpot(List<ChartLineSpot> chartLineSpots, DateTime xAxis,
      DateFormat dateFormat) {
    String xAxisShort = dateFormat.format(xAxis);
    List<ChartLineSpot> xAxisSpots = chartLineSpots
        .where((chartLineSpot) =>
            xAxisShort == dateFormat.format(chartLineSpot.date))
        .toList();

    ChartLineSpot avgSpotValue = xAxisSpots
        .reduce((min, element) => element.value < min.value ? element : min);
    return ChartLineSpot(xAxis, avgSpotValue.value);
  }

  List<ChartLineSpot> mean(List<ChartLineSpot> chartLineSpots) {
    chartLineSpots.sort((a, b) => a.date.compareTo(b.date));
    List<ChartLineSpot> meanChartLineSpots = [];
    for (int i = 0; i < chartLineSpots.length; i++) {
      double trailingSum = 0;
      for (int j = 0; j <= i; j++) {
        trailingSum += chartLineSpots[j].value;
      }
      meanChartLineSpots
          .add(ChartLineSpot(chartLineSpots[i].date, trailingSum / (i + 1)));
    }

    return meanChartLineSpots;
  }

  List<FlSpot> flSpots(
      List<ChartLineSpot> chartLineSpots, XAxisEnum xAxisEnum) {
    return chartLineSpots
        .map((chartLineSpot) => FlSpot(
            chartLineSpots.indexOf(chartLineSpot).toDouble(),
            chartLineSpot.value))
        .toList();
  }

  double maxX(XAxisEnum xAxisEnum) {
    return xAxisString(xAxisEnum).length.toDouble();
  }

  double maxY(List<LineChartBarData> lines) {
    double maxY = lines
        .reduce((maxLineChartBarData, lineChartBarData) =>
            maxLineY(maxLineChartBarData) > maxLineY(lineChartBarData)
                ? maxLineChartBarData
                : lineChartBarData)
        .spots
        .map((flSpot) => flSpot.y)
        .reduce(max);
    return maxY;
  }

  double maxLineY(LineChartBarData line) {
    return line.spots.map((flSpot) => flSpot.y).reduce(max);
  }
}
